<?php
namespace SIU\SQ\Lib\procesos;

use SIU\SQ\Lib\constantes;
use SIU\SQ\Lib\db\sql;
use SIU\SQ\Lib\entorno\configurador;
use SIU\SQ\Lib\errores\error;
use SIU\SQ\Lib\errores\error_negocio;
use SIU\SQ\Lib\lib;
use SIU\SQ\Lib\modelo\procesos_ejecuciones;

abstract class proceso
{
    protected $id_proceso;
    private $entorno;
    protected $estado_inicial;
    protected $verboso;
    protected $parametros = [];

    /** @var procesos_ejecuciones */
    protected $proceso_ejecucion;

    public function __construct($id_proceso, $entorno, $id_proceso_ejecucion = null, $estado_inicial = null)
    {
        $this->inicializar($estado_inicial);
        $this->entorno = $entorno;
        $this->id_proceso = $id_proceso;
        $this->cargar($id_proceso_ejecucion);
    }

    protected function inicializar($estado_inicial)
    {
        $this->estado_inicial 	= ($estado_inicial) ?  $estado_inicial : constantes::PROCESO_EJECUCION_ESTADO_INDEFINIDO;
        $this->verboso			= false;
    }

    private function cargar($id_proceso_ejecucion)
    {
        $this->datos = lib::catalogo_daos()->procesos()->get($this->id_proceso);
        if (!is_null($id_proceso_ejecucion)) {
            $this->proceso_ejecucion = lib::catalogo_modelo()->procesos_ejecuciones($id_proceso_ejecucion);
        } else {
            $this->proceso_ejecucion = lib::catalogo_modelo()->procesos_ejecuciones();
        }
        $this->proceso_ejecucion->set_descripcion($this->datos['descripcion']);
    }

    public function ejecutar()
    {
        if (!lib::es_modo_mantenimiento()) {
            $this->iniciar_registro_ejecucion();
            $this->entorno()->log()->info("INICIANDO PROCESO: [" . $this->datos['codigo'] . "] " . $this->datos['descripcion']);
            $this->ejecutar_especifico();
            $this->entorno()->log()->info("FINALIZANDO PROCESO: [" . $this->datos['codigo'] . "] " . $this->datos['descripcion']);
        } else {
            $config = new configurador();
            $dir = $config->get_config_dir();
            $proyecto = parse_ini_file($dir . "/../toba/proyecto.ini", true);
            $mensaje = $proyecto['proyecto']['mantenimiento_mensaje'] . PHP_EOL;
            echo $mensaje;
            $this->entorno->log()->warn($mensaje . " Intentando ejecutar el proceso: "
                . "([" . $this->datos['codigo'] . "] " . $this->datos['descripcion'] . ")");
        }
    }

    abstract public function ejecutar_especifico();

    public function entorno()
    {
        if (!isset($this->entorno)) {
            throw new error("PROCESO - El entorno no esta cargado");
        }
        return $this->entorno;
    }

    public function get_id_proceso()
    {
        return $this->id_proceso;
    }


    public function set_entorno($entorno)
    {
        $this->entorno = $entorno;
    }

    public function get_id_ejecucion()
    {
        return $this->proceso_ejecucion->get_id();
    }

    public function get_estado()
    {
        return $this->proceso_ejecucion->get_id_proceso_ejecucion_estado();
    }

    public function set_estado($estado)
    {
        $this->proceso_ejecucion->set_id_proceso_ejecucion_estado($estado);
    }

    public function set_total_a_procesar($total)
    {
        $this->proceso_ejecucion->set_total_a_procesar($total);
    }

    public function agregar_procesado_ok()
    {
        $this->proceso_ejecucion->set_procesados_ok($this->proceso_ejecucion->get_procesados_ok() + 1);
    }

    public function agregar_procesado_error()
    {
        $this->proceso_ejecucion->set_procesados_error($this->proceso_ejecucion->get_procesados_error() + 1);
    }

    public function es_verboso()
    {
        return $this->verboso;
    }

    public function set_verboso($verboso)
    {
        $this->verboso = $verboso;
    }

    public function iniciar_registro_ejecucion()
    {
        $this->proceso_ejecucion->set_id_proceso($this->id_proceso);
        $this->proceso_ejecucion->set_id_proceso_ejecucion_estado($this->estado_inicial);
        $this->proceso_ejecucion->guardar();
    }

    public function finalizar_registro_ejecucion($estado=null)
    {
        $estado_final = isset($estado)? $estado : $this->proceso_ejecucion->get_id_proceso_ejecucion_estado();

        $this->proceso_ejecucion->set_finalizado_en('now()');
        $this->proceso_ejecucion->set_id_proceso_ejecucion_estado($estado_final);
        $this->proceso_ejecucion->guardar();

        if ($estado_final == constantes::PROCESO_EJECUCION_ESTADO_ERROR) {
            $this->notificar_error();
        }
    }

    public function agregar_notificacion($mensaje, $severidad)
    {
        try {
            //------ Definir que se deber�a guardar en este campo...
            $tmp = debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS);
            unset($tmp[1]['file']);
            unset($tmp[1]['function']);
            unset($tmp[1]['type']);
            $componente = json_encode($tmp[1]);
            // Cuando es invocado desde la pantalla de toba la conexion a la db es LATIN1
            // por esa razón necesito convertir el mensaje en esa codificación.
            if ($this->es_encoding_latin1()) {
                $mensaje_utf8 = mb_convert_encoding($mensaje, 'ISO-8859-1', 'UTF-8');
            } else {
                $mensaje_utf8 = $mensaje;
            }

            //---------------------------------------------

            $datos = array( 'id_proceso_ejecucion'	=> $this->proceso_ejecucion->get_id(),
                'ocurrido_en'			=> 'now()',
                'severidad'				=> $severidad,
                'componente'			=> $componente,
                'texto'					=> $mensaje_utf8);
            $sql = sql::sql_array_a_insert('procesos_ejecuciones_log', $datos, $this->entorno->db());  // Se le pas� la DB porque internamente necesita la funci�n de quoteo de la clase db...
            $this->entorno->db()->ejecutar($sql);
        } catch (\Exception $ex) {
            $msg = "Error al grabar la tabla 'procesos_ejecuciones_log'. Id_proceso: " . $this->id_proceso . " - SQL: " . $sql;
            $this->entorno->log()->error(__METHOD__ . ": " . $msg . " - " . $ex->getMessage());
            throw new error($msg);
        }
    }

    public function get_notificaciones()
    {
        return lib::catalogo_daos()->procesos()->get_procesos_ejecuciones_log($this->proceso_ejecucion->get_id());
    }

    protected function es_encoding_latin1()
    {
        $db = $this->entorno->db();
        preg_match("/Client\sEncoding:\s(LATIN1)+;/", $db->get_pdo()->getAttribute(\PDO::ATTR_SERVER_INFO), $matches);
        if (empty($matches)) {
            return false;
        } else {
            return true;
        }
    }

    public function actualizar_registros_procesados()
    {
        // Consultar
        if ($this->proceso_ejecucion->get_id_proceso_ejecucion_estado() == constantes::PROCESO_EJECUCION_ESTADO_EN_ESPERA) {
            if ($this->proceso_ejecucion->get_total_a_procesar() == $this->proceso_ejecucion->get_procesados_ok()) {
                $this->proceso_ejecucion->set_id_proceso_ejecucion_estado(constantes::PROCESO_EJECUCION_ESTADO_OK);
            } else {
                if ($this->proceso_ejecucion->get_total_a_procesar() == ($this->proceso_ejecucion->get_procesados_ok() + $this->proceso_ejecucion->get_procesados_error())) {
                    $this->proceso_ejecucion->set_id_proceso_ejecucion_estado(constantes::PROCESO_EJECUCION_ESTADO_ERROR);
                }
            }
        }
        $this->proceso_ejecucion->guardar();
    }

    /**
     * Guarda un registro en notificaciones_clientes para que luego la levante el proceso de envio de emails
     */
    public function notificar_error()
    {
        $crear_notificacion = lib::parametros_sistema()->get_valor(constantes::PARAMETROS_SISTEMA_CODIGO_NOTIFICACION_ERROR_PROCESO);
        if ($crear_notificacion == 'S') {
            $param_usu_admin = lib::parametros_sistema()->get_valor(constantes::PARAMETROS_SISTEMA_CODIGO_USUARIO_ADMIN);
            $param_email_admin = lib::parametros_sistema()->get_valor(constantes::PARAMETROS_SISTEMA_CODIGO_EMAIL_ADMIN);
            if (empty($param_usu_admin) || empty($param_email_admin)) {
                $this->agregar_notificacion('El nombre o el email del usuario administrador no esta seteado en los parametros del sistema. '
                    . 'Para hacerlo puede usar el comando crear_usuario_admin', 'error');
            } else {
                $notificacion = lib::catalogo_modelo()->notificacion();
                $notificacion->set_usuario($param_usu_admin);
                $notificacion->set_notificado_a($param_email_admin);
                $notificacion->set_id_tipo_notificacion(constantes::TIPO_NOTIFICACION_ERROR_PROCESO);
                $notificacion->armar_tema($this->datos['descripcion']);
                $this->datos['id_ejecucion'] = $this->proceso_ejecucion->get_id();
                $notificacion->armar_mensaje($this->datos);
                $notificacion->guardar();
            }
        }
    }

    public function set_parametro($identificador, $valor)
    {
        $this->parametros[$identificador] = $valor;
    }

    public function set_descripcion_proceso_ejecucion($descripcion) {
        $this->proceso_ejecucion->set_descripcion($descripcion);
    }
    public function get_descripcion_proceso_ejecucion() {
        return $this->proceso_ejecucion->get_descripcion();
    }
}
