<?php

namespace SIU\SQ\Lib\util;

class validador_array
{
    protected $datos;
    protected $registros_ok;
    protected $registros_nook;
    protected $reglas;

    public function __construct()
    {
        $this->datos = [];
        $this->registros_ok = [];
        $this->registros_nook = [];
        $this->reglas = [];
        $this->rechazados = [];
    }

    public function set_datos($datos)
    {
        $this->datos = $datos;
    }

    public function get_datos()
    {
        return $this->datos;
    }
    
    public function get_rechazados()
    {
        return $this->rechazados;
    }

    public function set_reglas($v)
    {
        $this->reglas = $v;
    }

    /**
     * Realizo las validaciones sobre los datos cargados.
     *
     * @return array con resultado del proceso.
     */
    public function validar()
    {
        foreach ($this->datos as $d) {
            if ($this->procesar_registro($d)) {
                $this->registros_ok[] = $d;
            } else {
                $this->registros_nook[] = $d;
            }
        }

        return array(
            'registros_ok' => $this->registros_ok,
            'registros_nook' => $this->registros_nook
        );
    }

    /**
     * Recibo un registro y valido los campos
     *
     * @param Array $d Con el array del registro en si, realizo validaciones
     * de acuerdo a reglas recibidas, si ninguno devuelve false, sale true.
     *
     * @return boolean
     */
    protected function procesar_registro($d)
    {
        foreach ($this->reglas as $regla) {
            if (isset($regla['list'])) {
                $lista = $regla['list'];
            } else {
                $lista = [];
            }
            if (isset($d[$regla['name']])) {
                if (!$this->validar_tipo($d[$regla['name']], $regla['type'], $regla['length'], $regla['mandatory'], $lista)) {
                    $msg = "Error en campo: " . $regla['name'] . '. Tipo de dato: ' . $regla['type'] . '. ';
                    $msg.= $regla['mandatory'] ? 'Obligatorio. ' : 'No obligatorio. ';
                    $msg.= 'Valor obtenido: ' . $d[$regla['name']];
                    if (isset($regla['info_error'])) {
                        $msg.= '. Mensaje de error: ' . $regla['info_error'];
                    }
                    $this->rechazados[] = $msg;
                    return false;
                }
            } elseif ($regla['mandatory']) {
                $msg = "Error en campo: " . $regla['name'] . '. ';
                $msg.= 'Obligatorio. No se encuentra definido. ';
                $this->rechazados[] = $msg;
                return false;
            }
        }
        return true;
    }

    /**
     * Recibo un dato y le realizo las validaciones puntuales de acuerdo
     * al tipo de dato recibido.
     *
     * @param array $dato Es el dato a validar
     * @param string $tipo Es el tipo del dato que necesito convalidar
     * @param int $tamanio lenght maxima permitida para el dato
     * @param boolean $obligatorio valido que el dato tenga m�s de 0 char.
     *
     *
     * @return boolean
     */
    protected function validar_tipo($dato, $tipo, $tamanio = 100, $obligatorio = true, $valores_permitidos = [])
    {
        if ($obligatorio && strlen(trim($dato)) == 0) {
            return false;
        }

        if (strlen(trim($dato)) > $tamanio) {
            return false;
        }
        
        switch ($tipo) {
            
            case 'numeric':
                if (!is_numeric(trim($dato))) {
                    return false;
                }
                break;

            case 'int':
                if (!is_int(trim($dato))) {
                    return false;
                }
                break;

            case 'fecha':
                if (!$this->validateDate(trim($dato))) {
                    return false;
                }
                break;

            case 'string':
                break;

            case 'list':
                if (!in_array(trim($dato), $valores_permitidos)) {
                    return false;
                }
                break;

            case 'no-numeric':
                if (preg_match('~[0-9]+~', trim($dato))) {
                    return false;
                }
                break;

            case 'moneda':
                if (!preg_match('~^[0-9]+(\.[0-9]{1,2})?$~', trim($dato))) {
                    return false;
                }
                break;

            case 'fecha_pago':
                if (!$this->validarFechaPago(trim($dato))) {
                    return false;
                }
                break;

                default:
                    return false;
        }

        return true;
    }

    // internal helper
    // http://php.net/manual/es/function.checkdate.php#113205
    public function validateDate($date, $format = 'Y-m-d H:i:s')
    {
        $d = \DateTime::createFromFormat($format, $date);
        return $d && $d->format($format) == $date;
    }

    /**
     * Recibo una fecha y valido que tenga el formato correcto (d-m-Y).
     * También valido que esté entre el día 01-01-AñoActual y el día de hoy
     *
     * @param string $date Es la fecha a validar
     * @param string $format Es el formato con el que se validará
     *
     * @return \DateTime formateada o false si no pasa las validaciones
     */
    public function validarFechaPago($date, $format = 'd-m-Y')
    {
        $d = \DateTime::createFromFormat($format, $date);
        $fecha_hoy = new \DateTime('now');
        $fecha_minima = new \DateTime('01-12-' . ($fecha_hoy->format('Y') - 1));

        $formato_correcto = false;

        //"if($d)" valida que sea distinto de False. Puede ser false si el metodo createFromFormat falla.
        if ($d && $d >= $fecha_minima && $d <= $fecha_hoy) {
            $formato_correcto = $d;
        }

        return $formato_correcto;
    }
}
